import socket
import threading
import time


def get_serv(port):
    """
    Запрашивается название сервиса по порту. Если название получено,
    оно возвращается в функцию сканирования портов. Если нет, обрабатывается
    исключение и возвращается 'unassigned'.
    :param port: Порт для запроса имени сервиса. Целое число.
    :return: Возвращается сервис, если получен. В противном случае
    возвращается 'unassigned'.
    """
    try:
        return socket.getservbyport(port)
    except OSError:
        return 'unassigned'


class PortScan:
    """
    Класс для сканирования портов локального или удаленного компьютера.
    Из зависимостей требует библиотеку socket, а также наличие интернета,
    так как некоторые запросы обрабатываются с его помощью.
    """
    def __init__(self, target: str, port_counts: int):
        """
        При инициализации класса требуется передать ему адрес для сканирования и кол-во портов,
        которые требуется просканировать.
        :ip - обработанный адрес для сканирования, в котором,
        в случае невозможности получения адреса содержится None.
        :banners_port - словарь, в котором содержаться все открытые ip-порты и баннеры
        полученные от сервисов запущенных на них.
        :open_port - словарь, содержащий все открытые порты без баннеров.
        :param target: Полученный от пользователя адрес для сканирования. Тип: строка.
        :param port_counts: Полученное от пользователя кол-во портов для сканирования. Тип: целое число.
        """
        self.target = target
        self.ip = self.check_target()
        self.port_counts = int(port_counts)
        self.banners_port = dict()
        self.open_port = dict()

    def check_target(self):
        """
        Производиться обработка поступившего от пользователя адреса.
        Удаление http(https) для дальнейшего получения ip домена.
        Запрашивается ip-адрес, который возвращается из функции,
        если его удалось получить. Если нет, возвращается None.
        :return: Если не возникает исключения, возвращается ip-адрес,
        иначе возвращается None
        """
        if self.target.startswith("http"):
            self.target = self.target.split("/")[2]
        try:
            ip_domain = socket.gethostbyname(self.target)
            return ip_domain
        except socket.gaierror:
            return

    def scan_port(self, port: int):
        """
        В функции выполняется подключение к ip-адресу на полученный
        порт. Если подключение удалось, значит порт открыт. Производиться
        попытка получить баннер. Если баннер не получен, обрабатывается
        исключение, в котором идет запрос службы на порту у функции за
        пределами класса. Если получить службу не удалось, возвращается
        unassigned. Полученные значения добавляются в словари для
        последующей обработки. В словарь с баннерами добавляются значения,
        если баннер получен. В остальных случаях, в словарь открытых портов.
        :param port: Номер порта для сканирования. Целое число
        :return: Возвращается None. Служит для выхода из функции
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(5)
        if self.ip is None:
            return
        try:
            s.connect((self.ip, port))
            try:
                banner = s.recv(1024).decode().strip()
                if banner == '':
                    self.open_port.update({port: get_serv(port).upper()})
                else:
                    self.banners_port.update({port: banner})
            except OSError:
                self.open_port.update({port: get_serv(port).upper()})
            except UnicodeDecodeError:
                banner = s.recv(1024).strip()
                self.banners_port.update({port: banner})
        except (socket.timeout, ConnectionRefusedError, OSError):
            return

    def port_rotate(self):
        """
        Служит для перебора в цикле полученных от пользователя
        портов и передачи их в функцию сканирования, где производиться
        их обработка.
        :return: Ничего не возвращает.
        """
        threads = []
        for port in range(1, self.port_counts+1):
            t = threading.Thread(target=self.scan_port, kwargs={'port': port})
            t.daemon = True
            t.start()
            threads.append(t)
            time.sleep(0.02)
        for thread in threads:
            thread.join()
